@@ -62,9 +62,9 @@ impl Plugin {
     /// Panics if the world resource mutex could not be locked.
     pub unsafe fn instantiate(&self, sample_rate: f64) -> Result<Instance, InstantiateError> {
         let features = self.resources.features.lock().unwrap();
-        if features.min_and_max_block_length.is_none() {
-            return Err(InstantiateError::BlockLengthNotInitialized);
-        }
+        let (min_block_size, max_block_size) = features
+            .min_and_max_block_length
+            .ok_or(InstantiateError::BlockLengthNotInitialized)?;
         let instance = self
             .inner
             .instantiate(sample_rate, features.iter_features())
@@ -91,6 +91,8 @@ impl Plugin {
         }
         Ok(Instance {
             inner: instance.activate(),
+            min_block_size,
+            max_block_size,
             control_inputs,
             control_outputs,
             audio_inputs,
@@ -143,6 +145,8 @@ impl<'a> Debug for PortsDebug<'a> {
 /// An instance of a plugin that can process inputs and outputs.
 pub struct Instance {
     inner: lilv::instance::ActiveInstance,
+    min_block_size: usize,
+    max_block_size: usize,
     control_inputs: Vec<PortIndex>,
     control_outputs: Vec<PortIndex>,
     audio_inputs: Vec<PortIndex>,
@@ -196,6 +200,18 @@ impl Instance {
         CVOutputs: ExactSizeIterator + Iterator<Item = &'a mut [f32]>,
     {
         let sample_count = ports.sample_count;
+        if sample_count < self.min_block_size {
+            return Err(RunError::SampleCountTooSmall {
+                min_supported: self.min_block_size,
+                actual: sample_count,
+            });
+        }
+        if sample_count > self.max_block_size {
+            return Err(RunError::SampleCountTooLarge {
+                max_supported: self.max_block_size,
+                actual: sample_count,
+            });
+        }
         if ports.control_inputs.len() != self.control_inputs.len() {
             return Err(RunError::ControlInputsSizeMismatch {
                 expected: self.control_inputs.len(),
@@ -272,6 +288,7 @@ impl Instance {
             .atom_sequence_outputs
             .zip(self.atom_sequence_outputs.iter())
         {
+            data.clear_as_chunk();
             self.inner
                 .instance_mut()
                 .connect_port_mut(index.0, data.as_mut_ptr());
@@ -400,7 +417,7 @@ mod tests {
                 .instantiate(sample_rate)
                 .expect("Could not instantiate plugin.")
         };
-        let input = crate::event::LV2AtomSequence::new(1024);
+        let input = crate::event::LV2AtomSequence::new(&world, 1024);
         let params: Vec<f32> = plugin
             .ports_with_type(crate::PortType::ControlInput)
             .map(|p| p.default_value)
@@ -422,4 +439,56 @@ mod tests {
             })
         );
     }
+
+    #[test]
+    fn sample_count_smaller_than_supported_block_size_produces_error() {
+        let mut world = crate::World::new();
+        let supported_block_size = (512, 1024);
+        let lower_than_supported_block_size = 256;
+        world
+            .initialize_block_length(supported_block_size.0, supported_block_size.1)
+            .unwrap();
+        let plugin = world
+            .plugin_by_uri("http://drobilla.net/plugins/mda/EPiano")
+            .expect("Plugin not found.");
+        let mut instance = unsafe {
+            plugin
+                .instantiate(44100.0)
+                .expect("Could not instantiate plugin.")
+        };
+        let ports = crate::EmptyPortConnections::new(lower_than_supported_block_size);
+        assert_eq!(
+            unsafe { instance.run(ports) },
+            Err(crate::error::RunError::SampleCountTooSmall {
+                min_supported: 512,
+                actual: 256
+            })
+        );
+    }
+
+    #[test]
+    fn sample_count_larger_than_supported_block_size_produces_error() {
+        let mut world = crate::World::new();
+        let supported_block_size = (512, 1024);
+        let higher_than_supported_block_size = 2048;
+        world
+            .initialize_block_length(supported_block_size.0, supported_block_size.1)
+            .unwrap();
+        let plugin = world
+            .plugin_by_uri("http://drobilla.net/plugins/mda/EPiano")
+            .expect("Plugin not found.");
+        let mut instance = unsafe {
+            plugin
+                .instantiate(44100.0)
+                .expect("Could not instantiate plugin.")
+        };
+        let ports = crate::EmptyPortConnections::new(higher_than_supported_block_size);
+        assert_eq!(
+            unsafe { instance.run(ports) },
+            Err(crate::error::RunError::SampleCountTooLarge {
+                max_supported: 1024,
+                actual: 2048,
+            })
+        );
+    }
 }
